// Inigo Quilez - 2020
// The MIT License: Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions: The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

// This file is where all the EFI related stuff is, including the main entry point. All that happens here
// is that the graphics mode of the screen is queried, a backbuffer is allocated for rendering, a timer
// callback is created, the demo is inited, and we enter the render loop. I also initcialize the mlibc.h.cpp
// module which I created to provide basic libc functions (memset, malloc, fsqrt, fsin, fract, etc) that
// otherwise we no longer have available. It also helps decouple EFI form the demo stuff, which is in
// demo.h.cpp
//
// To install in an actual USB flash drive, copy the generated image/efi/boot/bootx64.efi file and copy it to your
// USB stick as /efi/boot/bootx64.efi.
//
// Some EFI documentation:
//    https://edk2-docs.gitbook.io/edk-ii-uefi-driver-writer-s-guide/5_uefi_services/readme.2/527_gettime
//    https://www.intel.com/content/dam/doc/guide/efi-driver-writers-v1-10-guide.pdf
//

#include <efi.h>

#include "mlibc.h"
#include "demo.h"

extern "C" int _fltused = 0;

static EFI_GRAPHICS_OUTPUT_PROTOCOL * iGfxInit( EFI_SYSTEM_TABLE *sys )
{
    EFI_GUID gEfiGraphicsOutputProtocolGuid = EFI_GRAPHICS_OUTPUT_PROTOCOL_GUID;

    // get from ConsoleOutHandle
    EFI_GRAPHICS_OUTPUT_PROTOCOL *gfx = nullptr;
    EFI_STATUS sta = sys->BootServices->HandleProtocol( sys->ConsoleOutHandle,&gEfiGraphicsOutputProtocolGuid,(VOID **) &gfx);
    if( EFI_ERROR(sta) )
    {
        // try locating directly
        sta = sys->BootServices->LocateProtocol( &gEfiGraphicsOutputProtocolGuid, NULL, (VOID **) &gfx);
        if( EFI_ERROR(sta) || gfx == NULL )
        {
            // try locating by handle
            UINTN num = 0;
            EFI_HANDLE *HandleBuffer = nullptr;
            sta = sys->BootServices->LocateHandleBuffer( ByProtocol, &gEfiGraphicsOutputProtocolGuid, nullptr, &num, &HandleBuffer);
            if( EFI_ERROR(sta) )
            {
                gfx = nullptr;
            }
            else
            {
                for( int i=0; i<num; i++ )
                {
                    sta = sys->BootServices->HandleProtocol( HandleBuffer[i],&gEfiGraphicsOutputProtocolGuid,(void**) &gfx);
                    if( !EFI_ERROR(sta) )
                        break;
                }
                sys->BootServices->FreePool(HandleBuffer);
            }
        }
    }

    return gfx;
}

static void iTimerHandler( EFI_EVENT evt, void *context )
{
    uint64_t *iTimeCounter = (uint64_t *)context;
    (*iTimeCounter) ++;
} 

static void *iMalloc( void *sys, uint64_t amount )
{
#if 1
    void *res = nullptr;
    EFI_STATUS status = ((EFI_SYSTEM_TABLE*)sys)->BootServices->AllocatePool( EfiBootServicesData, amount, &res);
    if( EFI_ERROR(status) )
        return nullptr;
    return res;
#else
    EFI_PHYSICAL_ADDRESS physicalRes;
    UINTN pages = EFI_SIZE_TO_PAGES (amount);
    EFI_STATUS status = ((EFI_SYSTEM_TABLE*)sys)->BootServices->AllocatePages( AllocateAnyPages, EfiBootServicesData, pages, &physicalRes);
    if( EFI_ERROR(status) )
        return nullptr;
    return (void *)(UINTN)physicalRes;
#endif
}
static void iFree( void *sys, void *ptr )
{
    ((EFI_SYSTEM_TABLE*)sys)->BootServices->FreePool(ptr);
    //((EFI_SYSTEM_TABLE*)sys)->BootServices->FreePages(ptr);

}
static void iMemcpy( void *sys, void *dst, void *src, uint64_t len )
{
    ((EFI_SYSTEM_TABLE*)sys)->BootServices->CopyMem(dst,src,len);
}
static void iMemset( void *sys, void *dst, uint8_t val, uint64_t len )
{
    ((EFI_SYSTEM_TABLE*)sys)->BootServices->SetMem(dst,len,val);
}
static void iPrintStr(void *sys, wchar_t *str )
{
    ((EFI_SYSTEM_TABLE*)sys)->ConOut->OutputString(((EFI_SYSTEM_TABLE*)sys)->ConOut, (CHAR16*)str);
}
static void iPrintInt(void *sys, int num )
{
    wchar_t str[16];
    str[0] = L'0' + (num/10000)%10;
    str[1] = L'0' + (num/1000)%10;
    str[2] = L'0' + (num/100)%10;
    str[3] = L'0' + (num/10)%10;
    str[4] = L'0' + (num/1)%10;
    str[5] = L'\r';
    str[6] = L'\n';
    str[7] = 0;
    iPrintStr(sys,str);
}

EFI_STATUS efi_main(EFI_HANDLE imageHandle, EFI_SYSTEM_TABLE *sys)
{
    //---------------------------
    // init graphics mode
    //---------------------------
    EFI_GRAPHICS_OUTPUT_PROTOCOL *gfx = iGfxInit(sys);
    if( gfx==nullptr )
    {
        iPrintStr( sys, L"Could not init graphics" );
        return EFI_LOAD_ERROR;
    }

    //---------------------------
    // allocate framebuffer
    //---------------------------
    const int xres = gfx->Mode->Info->HorizontalResolution;
    const int yres = gfx->Mode->Info->VerticalResolution;
    const int pitc = gfx->Mode->Info->PixelsPerScanLine;
    uint32_t *pixels = (uint32_t*)iMalloc(sys,4*xres*yres);
    if( !pixels )return EFI_LOAD_ERROR;

    #if 0
    const EFI_GRAPHICS_PIXEL_FORMAT format = gfx->Mode->Info->PixelFormat;
    if( format==PixelRedGreenBlueReserved8BitPerColor ) iPrintStr(sys, L"Format = BGRA\r\n" );
    if( format==PixelBlueGreenRedReserved8BitPerColor ) iPrintStr(sys, L"Format = RGBA\r\n" );
    if( format==PixelBitMask ) iPrintStr(sys, L"Format = PixelBitMask\r\n" );
    if( format==PixelBltOnly ) iPrintStr(sys, L"Format = PixelBltOnly\r\n" );
    iPrintStr(sys,L"xres: " );iPrintInt( sys,xres );
    iPrintStr(sys,L"yres: " );iPrintInt( sys,yres );
    iPrintStr(sys,L"pitc: " );iPrintInt( sys,pitc );
    #endif
    //---------------------------
    // setup timer
    //---------------------------
    uint64_t iTimeCounter = 0;
    EFI_EVENT evtTimer;
    EFI_STATUS status = sys->BootServices->CreateEvent( EFI_EVENT_TIMER | EFI_EVENT_NOTIFY_SIGNAL,EFI_TPL_CALLBACK, iTimerHandler,&iTimeCounter,&evtTimer ); 
    if( EFI_ERROR(status) ) return EFI_LOAD_ERROR;
    status = sys->BootServices->SetTimer( evtTimer,TimerPeriodic, 10000 );
    if( EFI_ERROR(status) ) return EFI_LOAD_ERROR;

    //---------------------------
    // init my mini libc module
    //---------------------------
    mlibc_init( sys, iMalloc, iFree, iMemcpy, iMemset, iPrintStr );

    //---------------------------
    // init demo
    //---------------------------
    if( !Demo_Init(xres, yres) ) return EFI_LOAD_ERROR;

    //---------------------------
    // reset keyboard
    //---------------------------
    sys->ConIn->Reset(sys->ConIn, FALSE);

    //---------------------------
    // render loop
    //---------------------------

    float iTime = 0.0f;
    bool done = false;
    while( !done )
    {
        // get time
        //sys->BootServices->Stall(4000); // 4000 microseconds = 4 miliseconds
        //const float iTime = float(iTimeCounter)/(float)(100);
        iTime += 1.0f/30.0f;

        // render demo
        Demo_Render( pixels, xres, yres, iTime );

        // blit should be more compatible than the memcpy with esoteric framebuffer pixel formats
        #if 1
        gfx->Blt(gfx, (EFI_GRAPHICS_OUTPUT_BLT_PIXEL*)pixels, EfiBltBufferToVideo, 0,0, 0,0, xres, yres, 0);
        #else
        char *dst = (char*)gfx->Mode->FrameBufferBase;
        char *src = (char*)pixels;
        for( int i=0; i<yres; i++ )
        {
            sys->BootServices->CopyMem( dst, src, xres*4 );
            src += xres*4;
            dst += pitc*4;
        }
        #endif
        // if ESC, exit
        EFI_INPUT_KEY key;
        status = sys->ConIn->ReadKeyStroke(sys->ConIn,&key);
        done = (key.ScanCode==SCAN_ESC);
    }
 
    //---------------------------
    // deinit demo
    //---------------------------
    Demo_DeInit();

    //---------------------------
    // free resources
    //---------------------------
    iFree(sys,pixels);
    sys->BootServices->SetTimer( evtTimer, TimerCancel, 0 ); 
    sys->BootServices->CloseEvent(evtTimer);

    //---------------------------
    // switch off machine
    //---------------------------
	sys->RuntimeServices->ResetSystem(EfiResetShutdown, EFI_SUCCESS, 0, nullptr);

	return EFI_SUCCESS;
}
